import collections
import functools
import importlib
import inspect
import itertools
import os
import re
import subprocess

from bdd_coder import exceptions
from bdd_coder import stock
from bdd_coder import features
from bdd_coder import tester

from bdd_coder.text_utils import (
    BASE_TESTER_NAME, extract_name, strip_lines, indent, make_method,
    make_class, make_doc, rstrip)

BDD_TESTER_PATH = 'tester.BddTester'


class FeatureClassCoder:
    def __init__(self, class_name, features_spec):
        self.class_name = class_name
        self.features_spec = features_spec

    @property
    def spec(self):
        return self.features_spec.features[self.class_name]

    @property
    def test_class_name(self):
        return self.features_spec.get_test_class_name(self.class_name)

    @property
    def scenario_method_defs(self):
        return [self.make_scenario_method_def(name, scenario_spec)
                for name, scenario_spec in self.spec['scenarios'].items()]

    @property
    def step_method_defs(self):
        steps = features.FeaturesSpec.get_all_steps(self.spec)

        return self.make_step_method_defs_for([s for s in steps if s.own])

    @property
    def class_body(self):
        head = '\n'.join(self.extra_class_attrs)

        return ('' if head else '\n') + '\n\n'.join(
            ([head] if head else []) + self.scenario_method_defs + self.step_method_defs)

    @property
    def extra_class_attrs(self):
        return [f'{name} = {value}' for name, value in self.spec['extra_class_attrs'].items()]

    @property
    def source(self):
        return make_class(
            self.test_class_name, self.spec['doc'], body=self.class_body, bases=self.bases)

    @property
    def bases(self):
        return [
            self.features_spec.get_test_class_name(b) for b in self.spec['bases']
        ] if self.spec['bases'] else [f'base.{BASE_TESTER_NAME}']

    @staticmethod
    def make_step_method_defs_for(steps_to_code):
        return [make_method(
            s.name, body=FeatureClassCoder.make_method_body(s.param_names, s.output_names),
            args_text=''.join([f', {n}' for n in s.param_names])
        ) for s in stock.list_drop_duplicates(steps_to_code, lambda s: s.name)]

    @staticmethod
    def make_scenario_method_def(name, scenario_spec, parameters=''):
        return make_method(
            ('' if scenario_spec['inherited'] else 'test_') + name,
            *scenario_spec['doc_lines'], decorators=('base.gherkin.scenario()',))

    @staticmethod
    def make_method_body(param_names, output_names):
        return (
            'return ' + ''.join(f"'{output}', " for output in output_names)
        ) if output_names else 'pass'


class PackageCoder:
    logs_file_name = 'bdd_runs.log'

    def __init__(self, specs_path='behaviour/specs', tests_path='',
                 test_module_name='stories', overwrite=False, logs_path=''):
        self.features_spec = features.FeaturesSpec.from_specs_dir(specs_path)
        self.tests_path = tests_path or os.path.join(os.path.dirname(specs_path), 'tests')
        self.logs_path = (
            logs_path or os.path.join(self.tests_path, self.logs_file_name)).rstrip('/')
        self.test_module_name = test_module_name
        self.overwrite = overwrite

    @property
    def story_class_defs(self):
        return [FeatureClassCoder(class_name, self.features_spec).source
                for class_name in self.features_spec.features]

    @property
    def aliases_def(self):
        dict_text = indent(
            '\n'.join([f"'{k}': '{v}'," for k, v in sorted(
                self.features_spec.aliases.items(), key=lambda it: it[0])]))

        return f'MAP = {{\n{dict_text}\n}}'

    @property
    def base_method_defs(self):
        return [make_method(name, args_text=', *args')
                for name in self.features_spec.base_methods]

    @property
    def base_class_def(self):
        return make_class(
            BASE_TESTER_NAME, bases=(BDD_TESTER_PATH,), decorators=('gherkin',),
            body='\n\n'.join(self.base_method_defs))

    def pytest(self):
        stock.Process('pytest', '-vv', self.tests_path).write()

    def write_aliases_module(self):
        with open(os.path.join(self.tests_path, 'aliases.py'), 'w') as aliases_py:
            aliases_py.write(rstrip(self.aliases_def) + '\n')

    @property
    def base_module_source(self):
        return '\n\n\n'.join([
            'from bdd_coder import decorators\n'
            'from bdd_coder import tester\n\n'
            'from . import aliases\n\n'
            f"gherkin = decorators.Gherkin(aliases.MAP, logs_path='{self.logs_path}')",
            self.base_class_def])

    def create_tester_package(self, run_pytest=False):
        exceptions.makedirs(self.tests_path, exist_ok=self.overwrite)

        with open(os.path.join(self.tests_path, '__init__.py'), 'w') as init_py:
            init_py.write("import pytest\n\n"
                          "pytest.register_assert_rewrite(f'{__name__}.base')\n")

        with open(os.path.join(self.tests_path, 'base.py'), 'w') as base_py:
            base_py.write(rstrip(self.base_module_source) + '\n')

        with open(os.path.join(self.tests_path, f'test_{self.test_module_name}.py'),
                  'w') as test_py:
            test_py.write(rstrip('\n\n\n'.join(
                ['from . import base'] + self.story_class_defs)) + '\n')

        self.write_aliases_module()

        if run_pytest:
            self.pytest()


class ModulePiece(stock.Repr):
    scenario_delimiter = '@base.gherkin.scenario'

    def __init__(self, text, name_regex=r'[A-Za-z0-9]+'):
        rtext = rstrip(text)
        match = re.match(
            fr'^(@(.+))?class ({name_regex})(.*?):\n(    """(.+?)""")?(.*)$',
            rtext, flags=re.DOTALL)

        if match:
            decorators, _, name, bases, _, doc, body = match.groups()
            self.name = extract_name(name)
            self.class_head = f'{name}{bases or ""}'
            self.decorators = decorators or ''
            self.doc = '\n'.join(strip_lines(doc.splitlines())) if doc else None
            self.body_head, self.scenarios, self.tail = self.split_class_body(body)
            self.match = True
        else:
            self.match = False
            self.name, self.text = f'piece{id(self)}', rtext

    def __str__(self):
        if not self.match:
            return self.text

        scenarios = '\n\n'.join(self.scenarios.values())
        scenarios_text = f'\n\n-- Scenarios:\n{scenarios}' if self.scenarios else ''

        return f'-- Head:\n{self.head}{scenarios_text}'

    @property
    def source(self):
        if not self.match:
            return self.text

        return rstrip('\n\n'.join(list(map(rstrip, itertools.chain(
            [self.head], self.scenarios.values(), self.tail)))))

    @property
    def head(self):
        doc = [indent(make_doc(
            *self.doc.splitlines()))] if self.doc else []
        body_head = [self.body_head] if self.body_head else []

        return '\n'.join([
            f'{self.decorators}class {self.class_head}:'] + doc + body_head)

    @classmethod
    def split_class_body(cls, text):
        pieces = iter(text.strip('\n').split(cls.scenario_delimiter))
        body_head = next(pieces).rstrip()
        scenarios, tail_pieces = collections.OrderedDict(), []

        for s in pieces:
            scenario_text, _, scenario_name, tail = cls.match_scenario_piece(s.strip())
            scenarios[scenario_name] = scenario_text

            if tail.strip():
                tail_pieces.append(tail.strip('\n'))

        return body_head, scenarios, tail_pieces

    @classmethod
    def match_scenario_piece(cls, text):
        scenario_code = f'    {cls.scenario_delimiter}{text}'
        match = re.match(
            r'^(    @base\.gherkin\.scenario\(.*\)\n    def (test_)?([^(]+)\(self\):\n'
            rf'{" "*8}"""\n.+?\n{" "*8}""")(.*)$',
            scenario_code, flags=re.DOTALL)

        if match is None:
            raise exceptions.ScenarioMismatchError(code=scenario_code)

        return match.groups()


class TestModule(stock.Repr):
    required_flake8_codes = [
        'E111', 'E301', 'E302', 'E303', 'E304', 'E701', 'E702', 'F999', 'W291', 'W293']

    def __init__(self, filename, *class_names):
        self.filename = filename
        self.tmp_filename = f'tmp_split_{id(self)}.py'
        self.flake8(filename)
        self.name_regex = r'|'.join(class_names)

        with open(filename) as py_file:
            self.pieces = self.split_module(rstrip(py_file.read()))

    def __str__(self):
        return '\n\n-----\n'.join([str(p) for p in self.pieces.values()])

    def __del__(self):
        if os.path.exists(self.tmp_filename):
            os.remove(self.tmp_filename)

    def transform(self, *mutations):
        for mutate in mutations:
            mutate(self.pieces)

        self.validate()

    def validate(self):
        with open(self.tmp_filename, 'w') as tmp_file:
            tmp_file.write(self.source)

        self.flake8(self.tmp_filename)

    def write(self):
        with open(self.filename, 'w') as py_file:
            py_file.write(self.source + '\n')

    @property
    def source(self):
        return rstrip('\n\n\n'.join([
            piece.source for piece in self.pieces.values()]))

    def flake8(cls, filename):
        try:
            subprocess.check_output([
                'flake8', '--select=' + ','.join(cls.required_flake8_codes), filename])
        except subprocess.CalledProcessError as error:
            raise exceptions.Flake8Error(error.stdout.decode())

    def split_module(self, source):
        return collections.OrderedDict([(mp.name, mp) for mp in (
            ModulePiece(piece, self.name_regex) for piece in source.split('\n\n\n'))])


class PackagePatcher(PackageCoder):
    default_specs_dir_name = 'specs'

    def __init__(self, test_module='behaviour.tests.test_stories', specs_path=''):
        """May raise `Flake8Error`"""
        self.base_tester, self.test_module = get_base_tester(test_module)
        self.new_specs = features.FeaturesSpec.from_specs_dir(specs_path or os.path.join(
            os.path.dirname(self.tests_path), self.default_specs_dir_name))
        old_specs = self.base_tester.features_spec()

        self.splits = {'base': TestModule(
            os.path.join(self.tests_path, 'base.py'), BASE_TESTER_NAME
        ), self.test_module_name: TestModule(
            os.path.join(self.tests_path, f'{self.test_module_name}.py'),
            *(old_specs.get_test_class_name(n) for n in old_specs.features))}

        new_classes = (
            set(self.new_specs.scenarios.values()) - set(old_specs.scenarios.values()))
        new_features = collections.OrderedDict([
            (cn, spec) for cn, spec in self.new_specs.features.items()
            if cn in new_classes])

        self.empty_classes = (
            set(old_specs.scenarios.values()) - set(self.new_specs.scenarios.values()))
        self.added_scenarios = {class_name: collections.OrderedDict(sorted(filter(
            lambda it: it[0] in set(self.new_specs.scenarios) - set(old_specs.scenarios),
            self.new_specs.features[class_name]['scenarios'].items()),
            key=lambda it: it[0])) for class_name in self.new_specs.scenarios.values()
            if class_name not in new_classes}
        self.removed_scenarios = {n: old_specs.scenarios[n] for n in (
            set(old_specs.scenarios) - set(self.new_specs.scenarios))}
        self.updated_scenarios = {n: self.new_specs.scenarios[n] for n in (
            set(old_specs.scenarios) & set(self.new_specs.scenarios)
        ) if self.new_specs.scenarios[n] == old_specs.scenarios[n] and (
            old_specs.features[old_specs.scenarios[n]]['scenarios'][n] !=
            self.new_specs.features[self.new_specs.scenarios[n]]['scenarios'][n])}

        self.features_spec = features.FeaturesSpec(new_features, (
            set(self.new_specs.base_methods) - set(old_specs.base_methods)
        ), {**old_specs.aliases, **self.new_specs.aliases})

    @property
    def tests_path(self):
        return os.path.dirname(self.test_module.__file__)

    @property
    def test_module_name(self):
        return self.test_module.__name__.rsplit('.', 1)[-1]

    def get_tester(self, class_name):
        return getattr(self.test_module, class_name)

    def patch_module(self, module_name, *mutations):
        self.splits[module_name].transform(*mutations)
        self.splits[module_name].write()

    def remove_scenarios(self, pieces):
        for name, class_name in self.removed_scenarios.items():
            del pieces[class_name].scenarios[name]

    def update_scenarios(self, pieces):
        for name, class_name in self.updated_scenarios.items():
            spec = self.new_specs.features[class_name]['scenarios'][name]
            pieces[class_name].scenarios[name] = indent(
                FeatureClassCoder.make_scenario_method_def(name, spec))

    def add_scenarios(self, pieces):
        for class_name, scenarios in self.added_scenarios.items():
            new_scenarios = {name: indent(
                FeatureClassCoder.make_scenario_method_def(name, spec))
                for name, spec in scenarios.items()}
            pieces[class_name].scenarios.update(new_scenarios)

    def add_new_stories(self, pieces):
        pieces.update({cp.name: cp for cp in (
            ModulePiece(text) for text in self.story_class_defs)})

    def sort_hierarchy(self, pieces):
        for class_name, piece in self.yield_sorted_pieces(pieces):
            pieces[class_name] = piece
            pieces.move_to_end(class_name)

        for class_name in self.empty_classes:
            self.update_class_head(
                class_name, f'{class_name}(base.{BASE_TESTER_NAME})', pieces)

    def yield_sorted_pieces(self, pieces):
        for name, bases in self.new_specs.class_bases:
            class_coder = self.get_class_coder(name)
            class_head = f'{class_coder.test_class_name}({", ".join(class_coder.bases)})'
            self.update_class_head(name, class_head, pieces)

            yield name, pieces[name]

    def get_class_coder(self, name):
        return FeatureClassCoder(name, self.new_specs)

    @staticmethod
    def update_class_head(name, class_head, pieces):
        pieces[name].class_head = class_head

    def update_docs(self, pieces):
        for name in self.new_specs.features.keys() - self.features_spec.features.keys():
            pieces[name].doc = self.new_specs.features[name]['doc']

    def add_new_steps(self, class_name, pieces):
        tester = self.get_tester(class_name)
        steps = self.new_specs.get_all_steps(self.new_specs.features[class_name])
        pieces[class_name].tail.extend(map(
            indent, FeatureClassCoder.make_step_method_defs_for(filter(
                lambda s: s.own and not hasattr(tester, s.name), steps))))

    def add_base_methods(self, pieces):
        pieces[BASE_TESTER_NAME].tail.extend(map(indent, self.base_method_defs))

    def patch(self, run_pytest=False):
        self.patch_module(
            self.test_module_name,
            self.remove_scenarios, self.update_docs, self.update_scenarios,
            self.add_scenarios, self.add_new_stories, self.sort_hierarchy, *[
                functools.partial(self.add_new_steps, subclass.__name__)
                for subclass in self.base_tester.subclasses_down()
                if subclass.__name__ in self.new_specs.features])
        self.patch_module('base', self.add_base_methods)
        self.write_aliases_module()

        if run_pytest:
            self.pytest()


def get_base_tester(test_module_path):
    try:
        test_module = importlib.import_module(test_module_path)
    except ModuleNotFoundError:
        raise exceptions.StoriesModuleNotFoundError(test_module=test_module_path)

    if not hasattr(test_module, 'base'):
        raise exceptions.BaseModuleNotFoundError(test_module=test_module)

    base_tester = {obj for name, obj in inspect.getmembers(test_module.base)
                   if inspect.isclass(obj) and tester.BddTester in obj.__bases__}

    if not len(base_tester) == 1:
        raise exceptions.BaseTesterNotFoundError(test_module=test_module, set=base_tester)

    return base_tester.pop(), test_module



import os
import sys
import traceback


class FeaturesSpecError(Exception):
    """Inconsistency in provided YAML specifications"""


class DocException(Exception):
    def __init__(self, *args, **kwargs):
        self.text = ' '.join(list(filter(None, map(str.strip, self.__doc__.format(
            *args, **kwargs).splitlines()))))

    def __str__(self):
        return self.text


class Flake8Error(Exception):
    """Some required flake8 tests failed"""


class OverwriteError(DocException):
    """Cannot overwrite {path} (--overwrite not set). {error}"""


def makedirs(path, exist_ok):
    try:
        os.makedirs(path, exist_ok=exist_ok)
    except OSError as error:
        raise OverwriteError(path=path, error=error)


def format_next_traceback():
    exc_type, exc_value, tb = sys.exc_info()
    tb_text = ''.join(traceback.format_list(traceback.extract_tb(tb.tb_next)))

    return exc_type, exc_value, (
        'Traceback (most recent call last):\n'
        f'{tb_text}{exc_type.__qualname__}: {exc_value}\n')


class ScenarioMismatchError(DocException):
    """Scenario code not understood: {code}..."""


class InconsistentClassStructure(DocException):
    """
    Expected class structure from docs does not match the defined one: {error}
    """


class WrongParametersError(DocException):
    """
    Invalid parameters at positions {positions} in scenario {name}:
    should be lists of length {length} (number of parameters declared in doc)
    """


class RedeclaredParametersError(DocException):
    """Redeclared parameter(s) {params}"""


class BaseTesterRetrievalError(DocException):
    """Raised in the base tester retrieval process"""


class StoriesModuleNotFoundError(BaseTesterRetrievalError):
    """Test module {test_module} not found"""


class BaseModuleNotFoundError(BaseTesterRetrievalError):
    """Test module {test_module} should have a `base` module imported"""


class BaseTesterNotFoundError(BaseTesterRetrievalError):
    """
    Imported base test module {test_module}.base should have a single
    BddTester subclass - found {set}
    """



from simple_cmd.decorators import ErrorsCommand

from bdd_coder.exceptions import (
    BaseTesterRetrievalError, FeaturesSpecError, InconsistentClassStructure,
    OverwriteError, Flake8Error, ScenarioMismatchError)

from bdd_coder import coders


@ErrorsCommand(FileNotFoundError, FeaturesSpecError, OverwriteError)
def make_blueprint(*,
                   specs_path: 'Directory containing the YAML specs' = 'behaviour/specs',
                   tests_path: 'Default: next to specs' = '',
                   test_module_name: 'Name for test_<name>.py' = 'stories',
                   overwrite=False,
                   run_pytest=False):
    coders.PackageCoder(
        specs_path=specs_path, tests_path=tests_path,
        test_module_name=test_module_name, overwrite=overwrite,
    ).create_tester_package(run_pytest=run_pytest)


@ErrorsCommand(BaseTesterRetrievalError, FeaturesSpecError, Flake8Error, ScenarioMismatchError)
def patch_blueprint(test_module: 'Passed to `importlib.import_module`',
                    specs_path: 'Directory to take new specs from. '
                    f'Default: {coders.PackagePatcher.default_specs_dir_name}/ '
                    'next to test package' = '', *, run_pytest=False):
    coders.PackagePatcher(test_module, specs_path).patch(run_pytest=run_pytest)


@ErrorsCommand(BaseTesterRetrievalError, OverwriteError, FeaturesSpecError,
               InconsistentClassStructure)
def make_yaml_specs(test_module: 'Passed to `importlib.import_module`',
                    specs_path: 'Will try to write the YAML files in here',
                    *, overwrite=False):
    base_tester, _ = coders.get_base_tester(test_module)
    features_spec = base_tester.features_spec(specs_path, overwrite)
    base_tester.validate_bases(features_spec)



import collections
import inspect
import os
import re
import shutil
import sys
import yaml

import pytest

from bdd_coder import exceptions
from bdd_coder import features
from bdd_coder import stock

from bdd_coder.text_utils import extract_name
from bdd_coder.text_utils import strip_lines
from bdd_coder.text_utils import to_sentence

from bdd_coder.exceptions import InconsistentClassStructure


class literal(str):
    """Employed to make nice YAML files"""


class YamlDumper:
    @staticmethod
    def dump_yaml(data, path):
        yaml.add_representer(collections.OrderedDict,
                             lambda dumper, data: dumper.represent_dict(data.items()))
        yaml.add_representer(literal, lambda dumper, data: dumper.represent_scalar(
            'tag:yaml.org,2002:str', data, style='|'))

        with open(path, 'w') as yml_file:
            yaml.dump(data, yml_file, default_flow_style=False)

    @classmethod
    def dump_yaml_aliases(cls, aliases, parent_dir):
        alias_lists = collections.defaultdict(list)

        for item in aliases.items():
            name, alias = map(to_sentence, item)
            alias_lists[alias].append(name)

        cls.dump_yaml(dict(alias_lists), os.path.join(parent_dir, 'aliases.yml'))


class BddTester(YamlDumper, stock.SubclassesMixin):
    """
    To be decorated with `Gherkin`
    """
    tmp_dir = '.tmp-specs'

    @classmethod
    def __init_subclass__(cls):
        if not hasattr(cls, 'gherkin'):
            return

        for name, scenario in cls.gherkin.scenarios[cls.__name__].items():
            for step in scenario.steps:
                try:
                    step_method = getattr(cls, step.name)
                except AttributeError:
                    raise exceptions.InconsistentClassStructure(
                        error=f'method {step.name} not found')

                step.method_qualname = step_method.__qualname__

                if step_method.__qualname__ in cls.gherkin:
                    step.doc_scenario = cls.gherkin[step_method.__qualname__]
                else:
                    setattr(cls, step.fixture_name, step(step_method))

        for scenario in filter(lambda s: not s.ready,
                               cls.gherkin.scenarios[cls.__name__].values()):
            setattr(cls, scenario.name, scenario(getattr(cls, scenario.name)))

    @classmethod
    def validate(cls):
        cls.validate_bases(cls.features_spec())

    @classmethod
    def features_spec(cls, parent_dir=None, overwrite=True):
        directory = parent_dir or cls.tmp_dir
        cls.dump_yaml_specs(directory, overwrite)

        try:
            return features.FeaturesSpec.from_specs_dir(directory)
        except exceptions.FeaturesSpecError as error:
            raise error
        finally:
            if parent_dir is None:
                shutil.rmtree(directory)

    @classmethod
    def validate_bases(cls, features_spec):
        spec_bases = collections.OrderedDict(features_spec.class_bases)
        cls_bases = collections.OrderedDict(
            (extract_name(c.__name__), b) for c, b in cls.subclasses_down().items())
        pair = stock.SetPair(spec_bases, cls_bases, lname='doc', rname='code')
        errors = []

        if not pair.symbol == '=':
            raise InconsistentClassStructure(error=f'Sets of class names differ: {repr(pair)}')

        for name in spec_bases:
            own_bases = set(cls_bases[name])
            own_bases.discard(cls)
            own_bases_names = {extract_name(b.__name__) for b in own_bases}

            if own_bases_names != spec_bases[name]:
                errors.append(f'bases {own_bases_names} declared in {name} do not '
                              f'match the specified ones {spec_bases[name]}')

        if errors:
            raise InconsistentClassStructure(error=', '.join(errors))

        sys.stdout.write('Test case hierarchy validated\n')

    @classmethod
    def dump_yaml_specs(cls, parent_dir, overwrite=False):
        exceptions.makedirs(parent_dir, exist_ok=overwrite)
        features_path = os.path.join(parent_dir, 'features')
        exceptions.makedirs(features_path, exist_ok=overwrite)

        cls.dump_yaml_aliases(cls.gherkin.aliases, parent_dir)

        for tester_subclass in cls.subclasses_down():
            tester_subclass.dump_yaml_feature(features_path)

        sys.stdout.write(f'Specification files generated in {parent_dir}\n')

    @classmethod
    def dump_yaml_feature(cls, parent_dir):
        name = '-'.join([s.lower() for s in cls.get_title().split()])
        cls.dump_yaml(cls.as_yaml(), os.path.join(parent_dir, f'{name}.yml'))

    @classmethod
    def as_yaml(cls):
        story = '\n'.join(map(str.strip, cls.__doc__.strip('\n ').splitlines()))
        scs = {to_sentence(re.sub('test_', '', name, 1)):
               strip_lines(getattr(cls, name).__doc__.splitlines())
               for name in cls.get_own_scenario_names()}

        return collections.OrderedDict([
            ('Title', cls.get_title()), ('Story', literal(story)), ('Scenarios', scs)
        ] + [(to_sentence(n), v) for n, v in cls.get_own_class_attrs().items()])

    @classmethod
    def get_title(cls):
        return re.sub(r'[A-Z]', lambda m: f' {m.group()}', extract_name(cls.__name__)).strip()

    @classmethod
    def get_own_scenario_names(cls):
        return list(cls.gherkin.scenarios[cls.__name__])

    @classmethod
    def get_own_class_attrs(cls):
        return dict(filter(lambda it: f'\n    {it[0]} = ' in inspect.getsource(cls),
                           inspect.getmembers(cls)))

    @classmethod
    def setup_class(cls):
        if cls.gherkin.validate:
            cls.gherkin.BddTester.validate()

    @pytest.fixture(autouse=True)
    def fixture_setup(self, request):
        self.gherkin.new_run(request.node.name, request.function.scenario)
        self.pytest_request = request
        self.gherkin.reset_outputs()

    @property
    def current_run(self):
        return self.gherkin.test_runs[self.pytest_request.node.name]

    def get_output(self, name, index=-1):
        return self.gherkin.outputs[name][index]



"""Common utils and constants"""
import os
import re

from bdd_coder import stock

BASE_TESTER_NAME = 'BddTester'

COMPLETION_MSG = 'All scenarios ran!'
OK, FAIL, PENDING, TO = '✔', '✖', '❓', '↦'
BOLD = {OK: '✅', FAIL: '❌'}

PARAM_REGEX = r'\$([a-zA-Z_]+)'
I_REGEX = r'\$\(([^\$]+)\)'
O_REGEX = r'`([^`\$]+)`'


class Style:
    end_mark = '\033[0m'

    @classmethod
    def purple(cls, text):
        return '\033[95m' + text + cls.end_mark

    @classmethod
    def dark_cyan(cls, text):
        return '\033[36m' + text + cls.end_mark

    @classmethod
    def cyan(cls, text):
        return '\033[96m' + text + cls.end_mark

    @classmethod
    def blue(cls, text):
        return '\033[94m' + text + cls.end_mark

    @classmethod
    def green(cls, text):
        return '\033[92m' + text + cls.end_mark

    @classmethod
    def yellow(cls, text):
        return '\033[93m' + text + cls.end_mark

    @classmethod
    def red(cls, text):
        return '\033[91m' + text + cls.end_mark

    @classmethod
    def bold(cls, text):
        return '\033[1m' + text + cls.end_mark

    @classmethod
    def underline(cls, text):
        return '\033[1m' + text + cls.end_mark


def to_sentence(name):
    return name.replace('_', ' ').capitalize()


def sentence_to_name(text):
    return '_'.join([re.sub(r'\W+', '', t).lower() for t in text.split()])


def strip_lines(lines):
    return list(filter(None, map(str.strip, lines)))


def extract_name(test_class_name):
    return test_class_name[4:] if test_class_name.startswith('Test') else test_class_name


def indent(text, tabs=1):
    newspace = ' '*4*tabs
    text = text.replace('\n', '\n' + newspace)

    return newspace + text


def make_doc(*lines):
    text = '\n'.join(line.strip() for line in lines)

    return f'"""\n{text}\n"""' if text else ''


def decorate(target, decorators):
    return '\n'.join([f'@{decorator}' for decorator in decorators] + [target])


def make_def_content(*doc_lines, body=''):
    return indent('\n'.join(([make_doc(*doc_lines)] if doc_lines else []) +
                            ([body] if body else []) or ['pass']))


def make_class_head(name, bases=(), decorators=()):
    inheritance = f'({", ".join(map(str.strip, bases))})' if bases else ''

    return decorate(f'class {name}{inheritance}', decorators)


def make_class(name, *doc_lines, bases=(), decorators=(), body=''):
    return '\n'.join([f'{make_class_head(name, bases, decorators)}:',
                      make_def_content(*doc_lines, body=body)])


def make_method(name, *doc_lines, args_text='', decorators=(), body=''):
    return '\n'.join([decorate(f'def {name}(self{args_text}):', decorators),
                      make_def_content(*doc_lines, body=body)])


def rstrip(text):
    return '\n'.join(list(map(str.rstrip, text.splitlines()))).lstrip('\n')


def assert_test_files_match(origin_dir, target_dir):
    py_file_names = ['__init__.py', 'aliases.py', 'base.py', 'test_stories.py']
    missing_in_origin = set(py_file_names) - set(os.listdir(origin_dir))
    missing_in_target = set(py_file_names) - set(os.listdir(target_dir))

    assert not missing_in_origin, f'Missing in {origin_dir}: {missing_in_origin}'
    assert not missing_in_target, f'Missing in {target_dir}: {missing_in_target}'

    diff_lines = {name: str(stock.Process(
        'diff', os.path.join(origin_dir, name), os.path.join(target_dir, name)
    )) for name in py_file_names}

    assert list(diff_lines.values()) == ['']*len(py_file_names), '\n'.join([
        f'Diff from {os.path.join(origin_dir, name)}:\n{diff}'
        for name, diff in diff_lines.items() if diff])



"""To be employed with `BddTester` and `BaseTestCase`"""
import collections
import datetime
import itertools
import functools
import logging
from logging.handlers import RotatingFileHandler

import pytest

from bdd_coder import exceptions
from bdd_coder.features import StepSpec
from bdd_coder import stock
from bdd_coder.text_utils import OK, FAIL, PENDING, TO, COMPLETION_MSG, BOLD, Style, indent


class StepRun(stock.Repr):
    def __init__(self, scenario_run, step, kwargs):
        self.scenario_run = scenario_run
        self.step = step
        self.kwargs = kwargs
        self.result = None

    @property
    def symbol(self):
        return getattr(self, '_StepRun__symbol', PENDING)

    @symbol.setter
    def symbol(self, value):
        self.__symbol = value
        self.__end_time = datetime.datetime.utcnow()

        if value == FAIL:
            self.scenario_run.symbol = FAIL
        elif value == OK and self.step.is_last:
            self.scenario_run.symbol = OK

    @property
    def end_time(self):
        return getattr(self, '_StepRun__end_time', None)

    @end_time.setter
    def end_time(self, value):
        raise AttributeError("'end_time' is read-only")

    def __str__(self):
        return (f'{self.end_time} {self.symbol} {self.step.method_qualname}'
                f'{self.step.format_parameters(**self.kwargs)} {self.formatted_result}')

    @property
    def formatted_result(self):
        if isinstance(self.result, tuple) and self.result:
            if self.symbol == OK:
                text = '\n'.join([f'    {repr(v)}' for v in self.result])

                return f'\n  {TO} {text.lstrip()}'

            if self.symbol == FAIL:
                return f'{TO} {self.result[2]}'

        return ''

    def log(self, **kwargs):
        lines = str(self).splitlines()
        lines[0] = f'├─{lines[0]}'
        lines[1:] = [f'|{line}' for line in lines[1:]]
        self.step.gherkin.logger.info('\n'.join(lines))


class ScenarioRun(stock.Repr):
    def __init__(self, test_run, scenario):
        super().__init__()
        self.test_run = test_run
        self.scenario = scenario
        self.runs = []

    def __iter__(self):
        yield from self.runs

    def __str__(self):
        qualname = self.scenario.qualname

        if self.symbol == FAIL:
            exc_type, exc_value, tb_text = self.result
            result_text = f' {TO} {exc_type.__name__}: {exc_value}'
        elif self.symbol == OK:
            result_text = f' {TO} {self.result}' if self.result else '.'

        return (f'{PENDING} {qualname}' if self.symbol == PENDING else
                f'{self.end_time} {BOLD[self.symbol]} {qualname}{result_text}')

    @property
    def result(self):
        return self.runs[-1].result if self.symbol != PENDING else ''

    @property
    def symbol(self):
        return getattr(self, '_ScenarioRun__symbol', PENDING)

    @symbol.setter
    def symbol(self, value):
        self.__symbol = value
        self.__end_time = datetime.datetime.utcnow()

        if value == FAIL and not self.scenario.is_test:
            self.test_run.symbol = FAIL

    @property
    def end_time(self):
        return getattr(self, '_ScenarioRun__end_time', None)

    @end_time.setter
    def end_time(self, value):
        raise AttributeError("'end_time' is read-only")

    def append_run(self, step, kwargs):
        self.runs.append(StepRun(self, step, kwargs))

    def log(self):
        self.scenario.gherkin.logger.info(f'└─{self}')


class TestRun(stock.Repr):
    def __init__(self, test_id, scenario):
        self.test_id = test_id
        self.scenario = scenario
        self.runs = collections.defaultdict(list)
        self.add_run(self.scenario)

    def __str__(self):
        return (f'{PENDING} ' if self.symbol == PENDING else
                f'{self.end_time} {BOLD[self.symbol]} ') + self.test_id

    def __iter__(self):
        yield from self.sorted_runs

    @property
    def sorted_runs(self):
        return sorted(itertools.chain(*self.runs.values()), key=lambda sr: sr.end_time)

    @property
    def scenario_run(self):
        return self.runs[self.scenario.name][-1]

    @property
    def symbol(self):
        return self.scenario_run.symbol

    @symbol.setter
    def symbol(self, value):
        self.scenario_run.symbol = value

    @property
    def end_time(self):
        return self.scenario_run.end_time

    @property
    def result(self):
        return self.scenario_run.result

    def add_run(self, scenario):
        self.runs[scenario.name].append(ScenarioRun(self, scenario))


class Step(StepSpec):
    def __init__(self, text, ordinal, scenario):
        super().__init__(text, ordinal, scenario.gherkin.aliases)
        self.scenario = scenario
        self.doc_scenario = None
        self.test_scenario = None
        self.is_last = False
        self.is_first = False
        self.method_qualname = ''

    @property
    def gherkin(self):
        return self.scenario.gherkin

    @property
    def fixture_param(self):
        if self.inputs:
            return [self.inputs[0] if len(self.inputs) == 1 else self.inputs]

    @property
    def fixture_name(self):
        return f'{self.name}{id(self)}'

    def __str__(self):
        return (f'Doc scenario {self.name}' if self.doc_scenario is not None
                else super().__str__())

    def __call__(self, step_method):
        @functools.wraps(step_method)
        def logger_step_method(tester, *args, **kwargs):
            if tester.current_run.symbol != PENDING:
                return

            if self.is_first and not self.scenario.is_test:
                tester.current_run.add_run(self.scenario)

            scenario_run = tester.current_run.runs[self.scenario.name][-1]
            scenario_run.append_run(self, kwargs)
            step_run = scenario_run.runs[-1]

            try:
                step_run.result = step_method(tester, *args, **kwargs)
            except Exception:
                step_run.result = exceptions.format_next_traceback()
                step_run.symbol = FAIL
            else:
                step_run.symbol = OK

                if isinstance(step_run.result, tuple):
                    for name, value in zip(self.output_names, step_run.result):
                        self.gherkin.outputs[name].append(value)

            step_run.log()

            if scenario_run.symbol != PENDING:
                scenario_run.log()

        return pytest.fixture(name=self.fixture_name, params=self.fixture_param)(
            logger_step_method)

    def format_parameters(self, **kwargs):
        if not kwargs and not self.inputs:
            return ''

        text = '\n'.join(([f'    {", ".join(self.inputs)}'] if self.inputs else []) +
                         [f'    {n} = {repr(v)}' for n, v in kwargs.items()])

        return f'\n{text}'


class Scenario(stock.Repr):
    def __init__(self, gherkin, *param_values):
        self.gherkin = gherkin
        self.param_values = param_values
        self.marked, self.ready = False, False

    def __str__(self):
        return f'{self.steps[0]}...{self.steps[-1]} params={self.param_names}'

    @property
    def name(self):
        return self.method.__name__

    @property
    def qualname(self):
        return self.method.__qualname__

    @property
    def param_names(self):
        names = []
        for name in itertools.chain(*(s.param_names for s in self.steps)):
            if name in names:
                raise exceptions.RedeclaredParametersError(params=name)
            else:
                names.append(name)
        return names

    def refine(self):
        fine_steps, param_ids, param_values = [], self.param_names, self.param_values
        wrong_values = [i for i, values in enumerate(param_values) if not (
            isinstance(values, list) and len(param_ids) == len(values))]

        if wrong_values:
            raise exceptions.WrongParametersError(
                name=self.name, positions=', '.join([str(i) for i in wrong_values]),
                length=len(param_ids))

        simple_steps = list(filter(lambda s: s.doc_scenario is None, self.steps))

        if simple_steps:
            simple_steps[0].is_first = True

            if simple_steps[-1] == self.steps[-1]:
                self.steps[-1].is_last = True

        for step in self.steps:
            if step.doc_scenario is None:
                fine_steps.append(step)
            else:
                finesteps, paramids, paramvalues = step.doc_scenario.refine()
                reused_ids = set(param_ids) & set(paramids)

                if reused_ids:
                    raise exceptions.RedeclaredParametersError(params=', '.join(reused_ids))

                param_ids.extend(paramids)
                fine_steps.extend(finesteps)

                param_values = (tuple(v1 + v2 for v1, v2 in zip(param_values, paramvalues))
                                if param_values else paramvalues)

        return fine_steps, param_ids, param_values

    def mark_method(self, method):
        self.steps = list(Step.generate_steps(method.__doc__.splitlines(), self))
        self.gherkin[method.__qualname__] = self
        self.is_test = method.__name__.startswith('test_')

        if self.is_test:
            return method

        @functools.wraps(method)
        def scenario_doc_method(tester, *args, **kwargs):
            raise AssertionError('Doc scenario method called')

        return scenario_doc_method

    def make_test_method(self, marked_method):
        fine_steps, param_ids, param_values = self.refine()

        @functools.wraps(marked_method)
        @pytest.mark.usefixtures(*(step.fixture_name for step in fine_steps))
        def scenario_test_method(tester, *args, **kwargs):
            __tracebackhide__ = True

            if tester.current_run.symbol == FAIL:
                pytest.fail(msg=tester.current_run.result[2], pytrace=False)

        if len(param_ids) == 1:
            param_values = [v[0] for v in param_values]

        if param_values:
            return pytest.mark.parametrize(
                ','.join(param_ids), param_values)(scenario_test_method)

        return scenario_test_method

    def __call__(self, method):
        if self.marked is False:
            self.method = self.mark_method(method)
            self.marked = True
        elif self.is_test and self.ready is False:
            self.method = self.make_test_method(method)
            self.ready = True

        self.method.scenario = self

        return self.method


class Gherkin(stock.Repr):
    def __init__(self, aliases, validate=True, **logging_kwds):
        self.reset_logger(**logging_kwds)
        self.reset_outputs()
        self.scenarios = collections.defaultdict(dict)
        self.aliases = aliases
        self.validate = validate
        self.test_runs = {}

    def __str__(self):
        return str(self.test_runs or self.scenarios)

    def __call__(self, BddTester):
        self.BddTester = BddTester
        BddTester.gherkin = self

        return BddTester

    def __contains__(self, scenario_qualname):
        class_name, method_name = scenario_qualname.split('.')

        return class_name in self.scenarios and method_name in self.scenarios[class_name]

    def __getitem__(self, scenario_qualname):
        class_name, method_name = scenario_qualname.split('.')

        return self.scenarios[class_name][method_name]

    def __setitem__(self, scenario_qualname, scenario_method):
        class_name, method_name = scenario_qualname.split('.')
        self.scenarios[class_name][method_name] = scenario_method

    def __iter__(self):
        for class_name in self.scenarios:
            yield from self.scenarios[class_name].values()

    def new_run(self, test_id, scenario):
        self.test_runs[test_id] = TestRun(test_id, scenario)
        self.logger.info('_'*26)

    def reset_logger(self, logs_path, maxBytes=100000, backupCount=10):
        self.logger = logging.getLogger('bdd_test_runs')
        self.logger.setLevel(level=logging.INFO)
        handler = RotatingFileHandler(logs_path, maxBytes=maxBytes, backupCount=backupCount)
        handler.setFormatter(logging.Formatter('%(message)s'))
        self.logger.handlers.clear()
        self.logger.addHandler(handler)

    def log(self, fail_if_pending=False):
        __tracebackhide__ = True
        runs = self.get_scenario_runs()
        self.logger.info('\n' + ''.join([
            f'  {len(runs[OK])}{BOLD[OK]}' if runs[OK] else '',
            f'  {len(runs[FAIL])}{BOLD[FAIL]}' if runs[FAIL] else '',
            f'  {len(runs[PENDING])}{PENDING}' if runs[PENDING] else f'  {COMPLETION_MSG}'
        ]) + '\n')

        if runs[FAIL]:
            self.logger.info('  ' + Style.bold('Scenario failures summary:'))

            for name, sruns in runs[FAIL].items():
                for r in sruns:
                    self.logger.info(indent(str(r)) + '\n')

        if runs[PENDING] and fail_if_pending:
            names = ', '.join(list(runs[FAIL]))
            pytest.fail(f'These scenarios did not run: {names}')

    def get_scenario_runs(self, symbols=(OK, FAIL, PENDING)):
        return {symbol: collections.OrderedDict(itertools.groupby(
            filter(lambda s: s.symbol == symbol, itertools.chain(*self.test_runs.values())),
            key=lambda s: s.scenario.name)) for symbol in symbols}

    def reset_outputs(self):
        self.outputs = collections.defaultdict(list)

    def scenario(self, *param_values):
        return Scenario(self, *param_values)



import collections
import itertools
import subprocess
import sys


class Repr:
    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'


class SubclassesMixin:
    @classmethod
    def subclasses_down(cls):
        clss, subclasses = [cls], []

        def chain_subclasses(classes):
            return list(itertools.chain(*map(lambda k: k.__subclasses__(), classes)))

        while clss:
            clss = chain_subclasses(clss)
            subclasses.extend(clss)

        return collections.OrderedDict([(sc, list(sc.__bases__)) for sc in subclasses])


class Process(subprocess.Popen):
    def __init__(self, *command, **kwargs):
        super().__init__(command, stdout=subprocess.PIPE, **kwargs)

    def __str__(self):
        return ''.join(list(self))

    def __iter__(self):
        line = self.next_stdout()

        while line:
            yield line

            line = self.next_stdout()

    def next_stdout(self):
        return self.stdout.readline().decode()

    def write(self, stream=sys.stdout):
        for line in self:
            stream.write(line)


class SetPair(Repr):
    def __init__(self, lset, rset, lname='l', rname='r'):
        self.lset, self.rset = set(lset), set(rset)
        self.lname, self.rname = lname, rname

    def __str__(self):
        return f'{self.lname} {self.symbol} {self.rname}: ' + ' | '.join(
            list(map(lambda s: '{' + ', '.join(sorted(map(repr, s))) + '}' if s else 'ø',
                     self.partition)))

    @property
    def partition(self):
        return [self.lset - self.rset, self.lset & self.rset, self.rset - self.lset]

    @property
    def symbol(self):
        parts = {}
        parts['l-r'], parts['l&r'], parts['r-l'] = self.partition

        if not parts['l&r']:
            return '⪥'
        elif not parts['l-r'] and not parts['r-l']:
            return '='
        elif parts['l-r'] and parts['r-l']:
            return '⪤'
        elif not parts['l-r']:
            return '⊂'
        elif not parts['r-l']:
            return '⊃'


def list_drop_duplicates(iterable, keylambda):
    elements, keys = [], []

    for e in filter(lambda x: keylambda(x) not in keys, iterable):
        keys.append(keylambda(e))
        elements.append(e)

    return elements



import configparser
import setuptools


ini = configparser.ConfigParser()
ini.read('version.ini')

with open('README.md', encoding='utf-8') as readme:
    long_description = readme.read()

tests_require = ['pytest-cov', 'freezegun', 'pytest-twisted', 'Twisted']

setuptools.setup(
    name=ini['version']['name'],
    version=ini['version']['value'],
    author='Daniel Farré Manzorro',
    author_email='d.farre.m@gmail.com',
    description='Gherkin language in class-based tests - test suite blueprinting',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/coleopter/bdd-coder',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Intended Audience :: Developers'],
    packages=setuptools.find_packages(),
    install_requires=['pyyaml', 'argparse', 'pytest', 'flake8', 'simple-cmd'],
    setup_requires=['setuptools', 'configparser'],
    tests_require=tests_require,
    extras_require={'dev': ['ipdb', 'ipython'], 'test': tests_require},
    entry_points={'console_scripts': [
        'bdd-blueprint=bdd_coder.commands:make_blueprint',
        'bdd-patch=bdd_coder.commands:patch_blueprint',
        'bdd-make-yaml-specs=bdd_coder.commands:make_yaml_specs']},
)



import collections
import copy
import itertools
import json
import os
import re
import yaml

from bdd_coder import exceptions
from bdd_coder import stock

from bdd_coder.text_utils import make_class_head
from bdd_coder.text_utils import sentence_to_name
from bdd_coder.text_utils import strip_lines
from bdd_coder.text_utils import I_REGEX, O_REGEX, PARAM_REGEX, TO

MAX_INHERITANCE_LEVEL = 100


class StepSpec(stock.Repr):
    @classmethod
    def generate_steps(cls, lines, *args, **kwargs):
        return (cls(line, i, *args, **kwargs) for i, line in enumerate(strip_lines(lines)))

    def __init__(self, text, ordinal, aliases):
        self.text = text.strip().split(maxsplit=1)[1].strip()
        self.ordinal = ordinal
        self.aliases = aliases
        self.own = False
        self.validate()

    def __str__(self):
        own = 'i' if self.own else 'o'
        output_names = ', '.join(self.output_names)
        param_names = ', '.join(self.param_names)

        return f'({own}) {self.name} [{param_names}] {TO} ({output_names})'

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (self.text, self.ordinal, self.own) == (other.text, other.ordinal, other.own)

    def validate(self):
        inames, onames = self.param_names, self.output_names
        has_repeated_inputs = len(set(inames)) < len(inames)
        has_repeated_outputs = len(set(onames)) < len(onames)

        if any([has_repeated_inputs, has_repeated_outputs]):
            raise exceptions.FeaturesSpecError(f'Repeated parameter names in {self}')

    @property
    def name(self):
        method = sentence_to_name(re.sub(I_REGEX, '', self.text))

        return self.aliases.get(method, method)

    @property
    def inputs(self):
        return re.findall(I_REGEX, self.text)

    @property
    def param_names(self):
        return re.findall(PARAM_REGEX, self.text)

    @property
    def output_names(self):
        return tuple(sentence_to_name(s) for s in re.findall(O_REGEX, self.text))


class FeaturesSpec(stock.Repr):
    @classmethod
    def from_specs_dir(cls, specs_path):
        """
        Constructs feature class specifications to be employed by the coders.
        Raises `FeaturesSpecError` for detected inconsistencies
        """
        aliases = cls.get_aliases(specs_path)
        prepared_specs = list(cls.yield_prepared_specs(specs_path, aliases))
        duplicates_errors = list(filter(None, (
            cls.check_if_duplicate_class_names(prepared_specs),
            cls.check_if_duplicate_scenarios(prepared_specs))))

        if duplicates_errors:
            raise exceptions.FeaturesSpecError('\n'.join(duplicates_errors))

        base_methods = set()
        features = cls.sets_to_lists(cls.sort(cls.simplify_bases(
            cls.check_if_cyclical_inheritance(cls.set_mro_bases(
                cls.prepare_inheritance_specs({
                    ft.pop('class_name'): ft for ft in prepared_specs
                }, aliases, base_methods))))))

        return cls(features, base_methods, aliases)

    def __init__(self, features, base_methods, aliases):
        self.features = features
        self.base_methods = sorted(base_methods)
        self.aliases = aliases

    def __str__(self):
        features = copy.deepcopy(self.features)

        for feature in features.values():
            for scenario in feature['scenarios'].values():
                scenario['steps'] = [str(step) for step in scenario['steps']]

        json_features = json.dumps(features, indent=4, ensure_ascii=False)
        bases = json.dumps(self.class_bases_text, indent=4)
        aliases = json.dumps(self.aliases, indent=4)
        base_methods = json.dumps(self.base_methods, indent=4)

        return '\n'.join([f'Class bases {bases}', f'Features {json_features}',
                          f'Aliases {aliases}', f'Base methods {base_methods}'])

    @property
    def scenarios(self):
        return self.get_scenarios(self.features)

    @property
    def class_bases(self):
        return list(map(lambda it: (it[0], set(it[1]['bases'])), self.features.items()))

    @property
    def class_bases_text(self):
        return list(map(lambda it: make_class_head(*it), self.class_bases))

    def get_test_class_name(self, class_name):
        return f'Test{class_name}' if self.is_test(class_name) else class_name

    def is_test(self, class_name):
        return not self.features[class_name]['inherited']

    @staticmethod
    def get_aliases(specs_path):
        with open(os.path.join(specs_path, 'aliases.yml')) as yml_file:
            yml_aliases = yaml.load(yml_file.read(), Loader=yaml.FullLoader)

        return dict(itertools.chain(*(zip(map(sentence_to_name, names), [
            sentence_to_name(alias)]*len(names))
            for alias, names in yml_aliases.items()))) if yml_aliases else {}

    @staticmethod
    def sets_to_lists(features):
        for feature_spec in features.values():
            feature_spec['bases'] = sorted(feature_spec['bases'])
            feature_spec['mro_bases'] = sorted(feature_spec['mro_bases'])

        return features

    @classmethod
    def prepare_inheritance_specs(cls, features, aliases, base_methods):
        for class_name, feature_spec in features.items():
            other_scenario_names = cls.get_scenarios(features, class_name)

            for step in cls.get_all_steps(feature_spec):
                if step.name in other_scenario_names:
                    other_class_name = other_scenario_names[step.name]
                    features[other_class_name]['inherited'] = True
                    features[other_class_name]['scenarios'][step.name]['inherited'] = True
                    feature_spec['mro_bases'].add(other_class_name)
                    feature_spec['bases'].add(other_class_name)
                elif step.name in feature_spec['scenarios']:
                    feature_spec['scenarios'][step.name]['inherited'] = True
                elif step.name in aliases.values():
                    base_methods.add(step.name)
                else:
                    step.own = True

        return features

    @classmethod
    def yield_prepared_specs(cls, specs_path, aliases):
        features_path = os.path.join(specs_path, 'features')

        for story_yml_name in os.listdir(features_path):
            with open(os.path.join(features_path, story_yml_name)) as feature_yml:
                yml_feature = yaml.load(feature_yml.read(), Loader=yaml.FullLoader)

            feature = {
                'class_name': cls.title_to_class_name(yml_feature.pop('Title')),
                'bases': set(), 'mro_bases': set(), 'inherited': False, 'scenarios': {
                    sentence_to_name(title): {
                        'title': title, 'inherited': False,  'doc_lines': lines,
                        'steps': tuple(StepSpec.generate_steps(lines, aliases))}
                    for title, lines in yml_feature.pop('Scenarios').items()},
                'doc': yml_feature.pop('Story')}
            feature['extra_class_attrs'] = {
                sentence_to_name(key): value for key, value in yml_feature.items()}

            yield feature

    @staticmethod
    def simplify_bases(features):
        for name, spec in filter(lambda it: len(it[1]['bases']) > 1, features.items()):
            bases = set(spec['bases'])

            for base_name in spec['bases']:
                bases -= bases & features[base_name]['bases']

            spec['bases'] = bases

        return features

    @staticmethod
    def check_if_cyclical_inheritance(features):
        for class_name, feature_spec in features.items():
            for base_class_name in feature_spec['mro_bases']:
                if class_name in features[base_class_name]['mro_bases']:
                    raise exceptions.FeaturesSpecError(
                        'Cyclical inheritance between {0} and {1}'.format(*sorted([
                            class_name, base_class_name])))

        return features

    @staticmethod
    def sort(features):
        """
        Sort (or try to sort) the features so that tester classes can be
        consistently defined.

        `MAX_INHERITANCE_LEVEL` is a limit for debugging, to prevent an
        infinite loop here, which should be forbidden by previous validation
        in the constructor
        """
        bases = {class_name: {
            'bases': spec['bases'], 'ordinal': 0 if not spec['bases'] else 1}
            for class_name, spec in features.items()}

        def get_of_level(ordinal):
            return {name for name in bases if bases[name]['ordinal'] == ordinal}

        level = 1

        while level < MAX_INHERITANCE_LEVEL:
            names = get_of_level(level)

            if not names:
                break

            level += 1

            for cn, bs in filter(lambda it: it[1]['bases'] & names, bases.items()):
                bs['ordinal'] = level
        else:
            raise AssertionError('Cannot sort tester classes to be defined!')

        return collections.OrderedDict(sorted(
            features.items(), key=lambda it: bases[it[0]]['ordinal']))

    @staticmethod
    def set_mro_bases(features):
        for name, spec in features.items():
            spec['mro_bases'].update(*(features[cn]['bases'] for cn in spec['bases']))
            spec['mro_bases'].discard(name)

        return features

    @staticmethod
    def check_if_duplicate_class_names(prepared_specs):
        repeated = list(map(
            lambda it: it[0], filter(lambda it: it[1] > 1, collections.Counter(
                [spec['class_name'] for spec in prepared_specs]).items())))

        if repeated:
            return f'Duplicate titles are not supported, {repeated}'

    @staticmethod
    def check_if_duplicate_scenarios(prepared_specs):
        scenarios = list(itertools.chain(*(
            [(nm, spec['class_name']) for nm in spec['scenarios']]
            for spec in prepared_specs)))
        repeated = dict(map(
            lambda it: (it[0], [cn for nm, cn in scenarios if nm == it[0]]),
            filter(lambda it: it[1] > 1,
                   collections.Counter([nm for nm, cn in scenarios]).items())))

        if repeated:
            return f'Repeated scenario names are not supported, {repeated}'

    @staticmethod
    def get_scenarios(features, *exclude):
        return {name: class_name for name, class_name in itertools.chain(*(
            [(nm, cn) for nm in spec['scenarios']]
            for cn, spec in features.items() if cn not in exclude))}

    @staticmethod
    def get_all_steps(feature_spec):
        return itertools.chain(*(
            sc['steps'] for sc in feature_spec['scenarios'].values()))

    @staticmethod
    def title_to_class_name(title):
        return ''.join(map(str.capitalize, title.split()))
